﻿using MesDataUpload.Core.Model;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using SQLite;

namespace MesDataUpload.Core
{
    public class Upload : INotifyPropertyChanged
    {
        private NLog.ILogger logger = NLog.LogManager.GetLogger("app");
        private static Upload instance=null;
        public static Upload Instance
        {
            get {
                if (instance == null)
                {
                    instance = new Upload();
                    instance.Init();
                    instance.Start();
                }
                return instance;
            }
        }
    
        public event PropertyChangedEventHandler PropertyChanged;

        public bool IsError { get; private set; }
        /// <summary>
        /// 错误信息
        /// </summary>
        public string ErrMsg { get; private set; }

        /// <summary>
        /// 状态信息
        /// </summary>
        public string Msg { get; private set; }

        public UploadParam Param { get; private set; } = new UploadParam();


        int IntervalMin  = 1;
        /// <summary>
        /// 剩余秒数
        /// </summary>
        public int SurplusSec { get; private set; }
        FbDbModel fbDb;
        ThkDbModel thkDb;

        public long LastScanDataId { get; private set; } = -1;
        public DateTime LastScanDataTime { get; private set; }

        public bool IsRunning { get; private set; }
        System.Threading.CancellationTokenSource cancellationTokenSource;
        public void Init()
        {
            Load();
            LoadLastId();
            this.PropertyChanged += Upload_PropertyChanged;
            SurplusSec = IntervalMin * 60;
        }

        private void Upload_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == "LastScanDataId")
            {
                SaveLastId();
            }
        }

        bool InitDb()
        {
            //自控数据库
            if (!System.IO.File.Exists(Param.FbDbPath))
            {
                ErrMsg = $"自控数据库文件不存在";
                return false;
            }
            try
            {
                fbDb = new FbDbModel(Param.FbDbPath);
                if (!fbDb.InitNoBuild())
                {
                    ErrMsg = $"自控数据库格式异常";
                    return false;
                }
            }
            catch (Exception e)
            {
                ErrMsg = e.Message;
                return false;
            }

            //测厚仪数据库
            if (!System.IO.File.Exists(Param.ThkDbPath))
            {
                ErrMsg = $"测厚仪数据库文件不存在";
                return false;
            }
            try
            {
                thkDb = new ThkDbModel(Param.ThkDbPath);
                if (!thkDb.InitNoBuild())
                {
                    ErrMsg = $"测厚仪数据库格式异常";
                    return false;
                }
            }
            catch (Exception e)
            {
                ErrMsg = e.Message;
                return false;
            }

            return true;
        }
        public void Start()
        {
            if (IsRunning)
                return;

            if (!InitDb())
            {
                IsError = true;
                return;
            }
            IsRunning = true;
            cancellationTokenSource = new System.Threading.CancellationTokenSource();


            Task.Factory.StartNew(() =>
            {
                while (!cancellationTokenSource.IsCancellationRequested)
                {
                    int totalsec = IntervalMin * 60-1;
                    for (int i = 0; i < totalsec; i++)
                    {
                        SurplusSec = totalsec - i;
                        try
                        {
                            Task.Delay(1000, cancellationTokenSource.Token).Wait();
                        }
                        catch (TaskCanceledException e)
                        {
                            //关闭
                            return;
                        }
                    }

                    try
                    {
                        IsError = !OnPoll();
                    }
                    catch (Exception e)
                    {

                        ErrMsg = e.Message;
                        IsError = true;
                        logger.Fatal(e);
                    }
                }
            });
        }



        bool OnPoll()
        {
            DateTime beginTime = DateTime.Now - TimeSpan.FromMinutes(IntervalMin*2);


            //读取 时间大于 (当前时间-轮询时间) 的 Db_ScanData 数据列
            var dataTable = thkDb.sqliteHelper.ExecuteReader($"select * from {thkDb.TbScanData.TableName} where Time >= { beginTime.ToStringOfSQLiteFieldType()} order by ID");
            if (dataTable.Rows.Count == 0)
            {
                ErrMsg = "异常，没有任何数据";
                return false;
            }

            List<Db_ScanData> db_ScanDatas = SQLite.SQLiteHelper.ToObjs<Db_ScanData>(dataTable);
            List<Lc_ScanData> lc_ScanDatas = Lc_AutoMapperProfile.Mapper.Map<List<Lc_ScanData>>(db_ScanDatas);
            if (lc_ScanDatas.Count == 0)
            {
                ErrMsg = "异常，不能翻译出任何数据";
                return false;
            }

            if (LastScanDataId > 0)
            {

                if (LastScanDataId > lc_ScanDatas.Last().ID)
                {
                    //之前的数据库被删除了

                }
                else
                {
                    //把数据列中 ID<= 上次上传ID 删除
                    lc_ScanDatas.RemoveAll(db => db.ID <= LastScanDataId);
                    if (lc_ScanDatas.Count() == 0)
                    {
                        //没有数据更新
                        Msg = "没有数据更新";
                        return true;
                    }
                }
            }
            LastScanDataId = lc_ScanDatas.Last().ID;

            string ids = "";
            foreach(var lc_ScanData in lc_ScanDatas)
            {
                if (!string.IsNullOrEmpty(ids))
                {
                    ids += ",";
                }
                ids += lc_ScanData.ProfileID.ToString();
            }

            //获取数据对应的 Db_Profile
            dataTable = thkDb.sqliteHelper.ExecuteReader($"select * from {thkDb.TbProfile.TableName} where ID in ({ids})");
            List<Db_Profile> db_Profiles = SQLite.SQLiteHelper.ToObjs<Db_Profile>(dataTable);

            DateTime scandata1stTime = lc_ScanDatas.First().Time;
            //获取样品
            dataTable = thkDb.sqliteHelper.ExecuteReader(
                $"select * from {thkDb.TbSampleData.TableName}" +
                $" where ID >= " +
                    $"(select ID from {thkDb.TbSampleData.TableName}" +
                    $" where Time <= {scandata1stTime.ToStringOfSQLiteFieldType()}" +
                    $" order by ID desc" +
                    $" limit 0,1)" +
                $" order by ID");
            List<Db_SampleData> db_Samples = SQLite.SQLiteHelper.ToObjs<Db_SampleData>(dataTable);
            List<Lc_SampleData> lc_SampleDatas = Lc_AutoMapperProfile.Mapper.Map<List<Lc_SampleData>>(db_Samples);



            //读 取 时间大于 (当前时间-轮询时间) 的自控 Db_FeedbackData 数据列
            dataTable = fbDb.sqliteHelper.ExecuteReader($"select * from {fbDb.TbFeedbackData.TableName} where Time >= { beginTime.ToStringOfSQLiteFieldType()} order by ID");
            List<Db_FeedbackData> db_FeedbackDatas = SQLite.SQLiteHelper.ToObjs<Db_FeedbackData>(dataTable);


            //生成 FLYMesDataLib.ModelClass
            List<FLYMesDataLib.ModelClass> modelClasses = new List<FLYMesDataLib.ModelClass>();
            
            foreach (Lc_ScanData lc_ScanData in lc_ScanDatas)
            {
                var profile = db_Profiles.Find(p => p.ID == lc_ScanData.ProfileID);
                if (profile == null)
                    continue;

                FLYMesDataLib.ModelClass mc = new FLYMesDataLib.ModelClass();

                mc.dept = Param.Dept;
                mc.cpxh = profile.PName;
                mc.scpc = profile.Batch;
                mc.jh = profile.Number;
                mc.sbbh = Param.Sbbh;
                mc.zfj = Param.Zfj;
                mc.jcwidth = (lc_ScanData.FilmEnd - lc_ScanData.FilmBegin).ToString();
                mc.tbwidth = mc.jcwidth;//TODO
                mc.dsm = profile.BSD == 2 ? "双面" : "单面";
                mc.targv = profile.Target.ToString("F1");
                mc.pcv = profile.Tolerance.ToString("F1");
                //数据库读的时间，都是 DateTimeKind.Unspecified
                //其实存进去时，是DateTimeKind.Local， 但ToString() 不完整
                //DateTime dtLc = DateTime.SpecifyKind(lc_ScanData.Time, DateTimeKind.Local);

                //服务器，要的是utc时间
                //mc.datetime = dtLc.ToUniversalTime();
                mc.datetime = DateTime.SpecifyKind(lc_ScanData.Time, DateTimeKind.Utc);

                var lc_SampleData = lc_SampleDatas.FindLast(sample => sample.Time <= lc_ScanData.Time);
                if (lc_SampleData != null)
                {
                    if (lc_SampleData.SampleADs != null)
                    {
                        if (lc_SampleData.SampleADs.Count() > 0)
                            mc.ypad0 = lc_SampleData.SampleADs[0].ToString();

                        if (lc_SampleData.SampleADs.Count() > 1)
                            mc.ypad1 = lc_SampleData.SampleADs[1].ToString();
                    }
                    
                }

                mc.xlbc = lc_ScanData.K.ToString("F3");
                mc.pybc = lc_ScanData.B.ToString("F2");

                if (lc_ScanData.Velocity > 2)
                    mc.valid = true.ToString();
                else
                    mc.valid = false.ToString();
                double avg = Average(lc_ScanData.Datas);
                if (double.IsNaN(avg))
                    mc.avg = "";
                else
                    mc.avg = avg.ToString("F2");

                
                var db_FeedbackData = db_FeedbackDatas.Find(fbd => fbd.ScanData_ID == lc_ScanData.ID);
                if (db_FeedbackData != null)
                {
                    mc.avg = db_FeedbackData.Avg.ToString("F2");
                    mc.zdj = db_FeedbackData.LeftDis.ToString();
                    mc.ydj = db_FeedbackData.RightDis.ToString();
                    mc.bs = db_FeedbackData.Pump.ToString("F2");
                }

                mc.data = lc_ScanData.Datas.Select(d => {
                    if (double.IsNaN(d))
                        return "";
                    else
                        return d.ToString("F2");
                }).ToList();
                modelClasses.Add(mc);
            }

            FLYMesDataLib.FLYMesData save = new FLYMesDataLib.FLYMesData();
            foreach (FLYMesDataLib.ModelClass mc in modelClasses)
            {
                bool bResult = save.SaveData(mc);
                if (!bResult)
                {
                    ErrMsg = "异常，FLYMesDataLib.FLYMesData.SaveData(mc) 失败";

                    return false;
                }
            }

            Msg = $"成功上传数据 最后一条数据时间为 {lc_ScanDatas.Last().Time}";
            return true;
        }
        public void Stop()
        {
            if (!IsRunning)
            {
                return;
            }
            cancellationTokenSource.Cancel();
            IsRunning = false;
            
        }

        double Average(IEnumerable<double> array)
        {
            var list2 = array.Where((d) =>
            {
                if (double.IsNaN(d))
                    return false;
                else
                    return true;
            });
            if (list2.Count() == 0)
                return double.NaN;

            return list2.Average();
        }
        public bool Load()
        {
            try
            {

                string json = File.ReadAllText("param.json");
                Newtonsoft.Json.JsonConvert.PopulateObject(json, Param);
                return true;
            }
            catch
            {
                return false;
            }
        }
        public bool Save()
        {
            try
            {
                File.WriteAllText("param.json", Newtonsoft.Json.JsonConvert.SerializeObject(Param, Newtonsoft.Json.Formatting.Indented));
                return true;
            }
            catch {
                return false;
            }
        }


        void LoadLastId()
        {
            try
            {

                string json = File.ReadAllText("lastid.json");
                LastScanDataId = Newtonsoft.Json.JsonConvert.DeserializeObject<long>(json);
            }
            catch
            {
                LastScanDataId = -1;
            }
        }
        void SaveLastId()
        {
            try
            {
                File.WriteAllText("lastid.json", Newtonsoft.Json.JsonConvert.SerializeObject(LastScanDataId));
            }
            catch
            {

            }
        }
    }
}
